from tfswin.model import SwinTransformer
from tfswin.model import SwinTransformerTiny224, SwinTransformerSmall224
from tfswin.model import SwinTransformerBase224, SwinTransformerBase384
from tfswin.model import SwinTransformerLarge224, SwinTransformerLarge384
from tfswin.model import SwinTransformerV2Tiny256, SwinTransformerV2Small256
from tfswin.model import SwinTransformerV2Base256, SwinTransformerV2Base384
from tfswin.model import SwinTransformerV2Large256, SwinTransformerV2Large384
from tfswin.prep import preprocess_input



import tensorflow as tf
from keras.testing_infra import test_combinations, test_utils
from tfswin.merge import PatchMerging


@test_combinations.run_all_keras_modes
class TestPatchMerging(test_combinations.TestCase):
    def test_layer(self):
        test_utils.layer_test(
            PatchMerging,
            kwargs={'swin_v2': False},
            input_shape=[2, 12, 12, 4],
            input_dtype='float32',
            expected_output_shape=[None, 6, 6, 8],
            expected_output_dtype='float32'
        )

        test_utils.layer_test(
            PatchMerging,
            kwargs={'swin_v2': True},
            input_shape=[2, 11, 13, 3],
            input_dtype='float32',
            expected_output_shape=[None, 6, 7, 6],
            expected_output_dtype='float32'
        )


if __name__ == '__main__':
    tf.test.main()



from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfswin.norm import LayerNorm


@register_keras_serializable(package='TFSwin')
class PatchEmbedding(layers.Layer):
    def __init__(self, patch_size, embed_dim, normalize, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.normalize = normalize

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.proj = layers.Conv2D(
            self.embed_dim, kernel_size=self.patch_size, strides=self.patch_size, padding='same', name='proj')

        if self.normalize:
            # noinspection PyAttributeOutsideInit
            self.norm = LayerNorm(name='norm')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = self.proj(inputs)

        if self.normalize:
            outputs = self.norm(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return self.proj.compute_output_shape(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({
            'patch_size': self.patch_size,
            'embed_dim': self.embed_dim,
            'normalize': self.normalize,
        })

        return config



import tensorflow as tf
from keras import initializers, layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFSwin')
class AbsoluteEmbedding(layers.Layer):
    def __init__(self, pretrain_size, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.pretrain_size = pretrain_size

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=4, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.embedding = self.add_weight(
            'embedding',
            shape=[1, self.pretrain_size, self.pretrain_size, channels],
            initializer=initializers.TruncatedNormal(stddev=0.02),
            trainable=True,
            dtype=self.dtype)

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        new_size = tf.shape(inputs)[1:3]
        embeddings = tf.image.resize(self.embedding, new_size, method=tf.image.ResizeMethod.BICUBIC)
        embeddings = tf.cast(embeddings, inputs.dtype)

        return inputs + embeddings

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({'pretrain_size': self.pretrain_size})

        return config



import numpy as np
import tensorflow as tf
from keras import layers
from keras.testing_infra import test_combinations
from keras.saving.object_registration import register_keras_serializable
from tfswin.swin import SwinBlock
from testing_utils import layer_multi_io_test


@register_keras_serializable('TFSwin')
class SwinBlockSqueeze(SwinBlock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_spec = [
            layers.InputSpec(ndim=4), layers.InputSpec(ndim=1, dtype='int32'), layers.InputSpec(ndim=1, dtype='int32'),
            layers.InputSpec(ndim=2, dtype='int32'), layers.InputSpec(ndim=5)]

    def call(self, inputs, **kwargs):
        inputs, shift_size, window_size, relative_index, attention_mask = inputs

        return super().call([inputs, shift_size[0], window_size[0], relative_index[0], attention_mask], **kwargs)


@test_combinations.run_all_keras_modes
class TestSwinBlock(test_combinations.TestCase):
    def test_layer(self):
        inputs = 10 * np.random.random((1, 7, 7, 96)) - 0.5
        shift3 = np.array([3], 'int32')
        shift5 = np.array([5], 'int32')
        window = np.array([7], 'int32')
        index = np.zeros([1, 7 ** 4], 'int32')
        masks = 10 * np.random.random((1, 1, 1, 49, 49)) - 0.5

        layer_multi_io_test(
            SwinBlockSqueeze,
            kwargs={'num_heads': 24, 'mlp_ratio': 4., 'qkv_bias': True, 'qk_scale': None, 'drop': 0., 'attn_drop': 0.,
                    'path_drop': 0.20000000298023224, 'window_pretrain': 7, 'swin_v2': False},
            input_datas=[inputs, shift3, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, 7, 7, 96)],
            expected_output_dtypes=['float32']
        )
        layer_multi_io_test(
            SwinBlockSqueeze,
            kwargs={'num_heads': 24, 'mlp_ratio': 4., 'qkv_bias': True, 'qk_scale': None, 'drop': 0., 'attn_drop': 0.,
                    'path_drop': 0.20000000298023224, 'window_pretrain': 7, 'swin_v2': False},
            input_datas=[inputs, shift5, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, 7, 7, 96)],
            expected_output_dtypes=['float32']
        )

        layer_multi_io_test(
            SwinBlockSqueeze,
            kwargs={'num_heads': 24, 'mlp_ratio': 4., 'qkv_bias': True, 'qk_scale': None, 'drop': 0., 'attn_drop': 0.,
                    'path_drop': 0.20000000298023224, 'window_pretrain': 0, 'swin_v2': True},
            input_datas=[inputs, shift3, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, 7, 7, 96)],
            expected_output_dtypes=['float32']
        )
        layer_multi_io_test(
            SwinBlockSqueeze,
            kwargs={'num_heads': 24, 'mlp_ratio': 4., 'qkv_bias': True, 'qk_scale': None, 'drop': 0., 'attn_drop': 0.,
                    'path_drop': 0.20000000298023224, 'window_pretrain': 4, 'swin_v2': True},
            input_datas=[inputs, shift5, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, 7, 7, 96)],
            expected_output_dtypes=['float32']
        )


if __name__ == '__main__':
    tf.test.main()



import numpy as np
import tensorflow as tf
from keras import layers
from keras.testing_infra import test_combinations
from keras.saving.object_registration import register_keras_serializable
from tfswin.winatt import WindowAttention
from testing_utils import layer_multi_io_test


@register_keras_serializable('TFSwin')
class WindowAttentionSqueeze(WindowAttention):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_spec = [
            layers.InputSpec(ndim=4), layers.InputSpec(ndim=1, dtype='int32'), layers.InputSpec(ndim=2, dtype='int32'),
            layers.InputSpec(ndim=5)]

    def call(self, inputs, **kwargs):
        inputs, window_size, relative_index, attention_mask = inputs

        return super().call([inputs, window_size[0], relative_index[0], attention_mask], **kwargs)


@test_combinations.run_all_keras_modes
class TestWindowAttention(test_combinations.TestCase):
    def test_layer(self):
        inputs = 10 * np.random.random((1, 7, 7, 96)) - 0.5
        window = np.array([7], 'int32')
        index = np.zeros([1, 7 ** 4], 'int32')
        masks = 10 * np.random.random((1, 1, 1, 49, 49)) - 0.5

        layer_multi_io_test(
            WindowAttentionSqueeze,
            kwargs={'num_heads': 3, 'qkv_bias': True, 'qk_scale': None, 'attn_drop': 0., 'proj_drop': 0.,
                    'window_pretrain': 7, 'swin_v2': False},
            input_datas=[inputs, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, None, None, 96)],
            expected_output_dtypes=['float32']
        )
        layer_multi_io_test(
            WindowAttentionSqueeze,
            kwargs={'num_heads': 3, 'qkv_bias': True, 'qk_scale': None, 'attn_drop': 0., 'proj_drop': 0.,
                    'window_pretrain': 0, 'swin_v2': False},
            input_datas=[inputs, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, None, None, 96)],
            expected_output_dtypes=['float32']
        )

        layer_multi_io_test(
            WindowAttentionSqueeze,
            kwargs={'num_heads': 3, 'qkv_bias': True, 'qk_scale': None, 'attn_drop': 0., 'proj_drop': 0.,
                    'window_pretrain': 0, 'swin_v2': True},
            input_datas=[inputs, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, None, None, 96)],
            expected_output_dtypes=['float32']
        )
        layer_multi_io_test(
            WindowAttentionSqueeze,
            kwargs={'num_heads': 3, 'qkv_bias': True, 'qk_scale': None, 'attn_drop': 0., 'proj_drop': 0.,
                    'window_pretrain': 8, 'swin_v2': True},
            input_datas=[inputs, window, index, masks],
            input_dtypes=['float32', 'int32', 'int32', 'float32'],
            expected_output_shapes=[(None, None, None, 96)],
            expected_output_dtypes=['float32']
        )


if __name__ == '__main__':
    tf.test.main()



import numpy as np
import tensorflow as tf
from keras import layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfswin.swin import SwinBlock
from tfswin.window import window_partition


@register_keras_serializable(package='TFSwin')
class BasicLayer(layers.Layer):
    def __init__(self, depth, num_heads, window_size, mlp_ratio=4., qkv_bias=True, qk_scale=None,
                 drop=0., attn_drop=0., path_drop=0., window_pretrain=0, swin_v2=False, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.depth = depth
        self.num_heads = num_heads
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop = drop
        self.attn_drop = attn_drop
        self.path_drop = path_drop
        self.window_pretrain = window_pretrain
        self.swin_v2 = swin_v2

        self.shift_size = self.window_size // 2

    @shape_type_conversion
    def build(self, input_shape):
        path_drop = self.path_drop
        if not isinstance(self.path_drop, (list, tuple)):
            path_drop = [self.path_drop] * self.depth

        # noinspection PyAttributeOutsideInit
        self.blocks = [
            SwinBlock(num_heads=self.num_heads, mlp_ratio=self.mlp_ratio, qkv_bias=self.qkv_bias,
                      qk_scale=self.qk_scale, drop=self.drop, attn_drop=self.attn_drop, path_drop=path_drop[i],
                      window_pretrain=self.window_pretrain, swin_v2=self.swin_v2, name=f'blocks.{i}')
            for i in range(self.depth)]

        super().build(input_shape)

    def shift_window(self, height, width):
        min_size = tf.minimum(height, width)
        with_shift = min_size > self.window_size
        shift_size = self.shift_size * tf.cast(with_shift, min_size.dtype)
        window_size = tf.minimum(self.window_size, min_size)

        return shift_size, window_size

    def relative_index(self, window_size):
        offset = tf.range(window_size)
        offset = tf.stack(tf.meshgrid(offset, offset, indexing='ij'), axis=0)
        offset = tf.reshape(offset, [2, -1])
        offset = offset[:, :, None] - offset[:, None]

        index = offset + (window_size - 1)
        index = index[0] * (2 * window_size - 1) + index[1]
        index = tf.reshape(index, [-1])

        return index

    def attention_mask(self, height, width, shift_size, window_size):
        padded_height = tf.cast(tf.math.ceil(height / window_size), 'int32') * window_size
        padded_width = tf.cast(tf.math.ceil(width / window_size), 'int32') * window_size

        last_repeats = [window_size - shift_size, shift_size]

        shift_mask = np.arange(9, dtype='int32').reshape((3, 3))
        shift_mask = tf.repeat(shift_mask, [padded_height - window_size] + last_repeats, axis=1)
        shift_mask = tf.repeat(shift_mask, [padded_width - window_size] + last_repeats, axis=0)
        shift_mask = shift_mask[None, ..., None]
        shift_windows = window_partition(shift_mask, padded_height, padded_width, window_size, 'int32')
        shift_windows = tf.squeeze(shift_windows, axis=-1)
        shift_windows = shift_windows[:, None] - shift_windows[:, :, None]

        pad_mask = tf.ones((1, height, width, 1), dtype='int32')
        pad_mask = tf.pad(pad_mask, [[0, 0], [0, padded_height - height], [0, padded_width - width], [0, 0]])
        pad_windows = window_partition(pad_mask, padded_height, padded_width, window_size, 'int32')
        pad_windows = tf.squeeze(pad_windows, axis=-1)
        pad_windows = pad_windows[:, None] - pad_windows[:, :, None]

        attn_mask = tf.where((shift_windows == 0) & (pad_windows == 0), 0., -100.)
        attn_mask = tf.cast(attn_mask, self.compute_dtype)
        attn_mask = attn_mask[None, :, None]

        return attn_mask

    def call(self, inputs, *args, **kwargs):
        height, width = tf.unstack(tf.shape(inputs)[1:3])

        shift_size, window_size = self.shift_window(height, width)
        relative_index = self.relative_index(window_size)
        shift_mask = self.attention_mask(height, width, shift_size, window_size)
        identity_mask = self.attention_mask(height, width, 0, window_size)

        outputs = inputs
        for i, b in enumerate(self.blocks):
            current_shift = shift_size if i % 2 else 0
            current_mask = shift_mask if i % 2 else identity_mask
            outputs = b([outputs, current_shift, window_size, relative_index, current_mask])

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            'depth': self.depth,
            'num_heads': self.num_heads,
            'window_size': self.window_size,
            'mlp_ratio': self.mlp_ratio,
            'qkv_bias': self.qkv_bias,
            'qk_scale': self.qk_scale,
            'drop': self.drop,
            'attn_drop': self.attn_drop,
            'path_drop': self.path_drop,
            'window_pretrain': self.window_pretrain,
            'swin_v2': self.swin_v2
        })

        return config



from keras import layers
from keras.saving.object_registration import register_keras_serializable


@register_keras_serializable(package='TFSwin')
class LayerNorm(layers.LayerNormalization):
    # Overload to use fused implementation

    def __init__(self, epsilon=1.001e-5, **kwargs):
        super().__init__(epsilon=epsilon, **kwargs)

    def _fused_can_be_used(self, ndims):
        if 'float32' != self.dtype:
            raise ValueError(
                f'Fused layer normalization is only supported when the variables dtype is '
                f'float32. Got dtype: {self.dtype}.')

        if self._compute_dtype not in ('float16', 'float16', 'float32', None):
            raise ValueError(
                f'Fused layer normalization is only supported when the compute dtype is '
                f'float16, bfloat16, or float32. Got dtype: {self._compute_dtype}.')

        if self.epsilon < 1.001e-5:
            raise ValueError(
                f'Fused layer normalization is not supported for epsilon {self.epsilon} (<1.001e-5).')

        axis = sorted(self.axis)
        if axis[-1] != ndims - 1 or axis[-1] - axis[0] != len(axis) - 1:
            raise ValueError(
                f'Fused layer normalization is not supported for axis {self.axis} and inputs rank {ndims}.')

        return True



import numpy as np
import tensorflow as tf
from keras import initializers, layers
from keras.saving.object_registration import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfswin.window import window_partition_fused, window_reverse_fused


@register_keras_serializable(package='TFSwin')
class WindowAttention(layers.Layer):
    def __init__(self, num_heads, qkv_bias=True, qk_scale=None, attn_drop=0., proj_drop=0.,
                 window_pretrain=0, swin_v2=False, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = [
            layers.InputSpec(ndim=4), layers.InputSpec(ndim=0, dtype='int32'), layers.InputSpec(ndim=1, dtype='int32'),
            layers.InputSpec(ndim=5)]

        self.num_heads = num_heads
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.attn_drop = attn_drop
        self.proj_drop = proj_drop
        self.window_pretrain = window_pretrain
        self.swin_v2 = swin_v2

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.channels = input_shape[0][-1]
        if self.channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')

        qkv_bias = not self.swin_v2 and self.qkv_bias
        # noinspection PyAttributeOutsideInit
        self.qkv = layers.Dense(self.channels * 3, use_bias=qkv_bias, name='qkv')

        if self.swin_v2:
            # noinspection PyAttributeOutsideInit
            self.scale = self.add_weight(
                'logit_scale',
                shape=[self.num_heads, 1, 1],
                initializer=initializers.Constant(np.log(10.)),
                trainable=True,
                dtype=self.dtype)

            # noinspection PyAttributeOutsideInit
            self.cpb0 = layers.Dense(512, activation='relu', name='cpb_mlp.0')
            self.cpb1 = layers.Dense(self.num_heads, activation='sigmoid', use_bias=False, name=f'cpb_mlp.2')

            # noinspection PyAttributeOutsideInit
            self.q_bias = None
            # noinspection PyAttributeOutsideInit
            self.v_bias = None
            if self.qkv_bias:
                self.q_bias = self.add_weight(
                    'q_bias',
                    shape=[self.channels],
                    initializer='zeros',
                    trainable=True,
                    dtype=self.dtype)
                self.v_bias = self.add_weight(
                    'v_bias',
                    shape=[self.channels],
                    initializer='zeros',
                    trainable=True,
                    dtype=self.dtype)
        else:
            # noinspection PyAttributeOutsideInit
            self.scale = self.qk_scale or (self.channels // self.num_heads) ** -0.5

            # noinspection PyAttributeOutsideInit
            self.relative_bias = self.add_weight(
                'relative_position_bias_table',
                shape=[(2 * self.window_pretrain - 1) ** 2, self.num_heads],
                initializer=initializers.TruncatedNormal(stddev=0.02),
                trainable=True,
                dtype=self.dtype)

        # noinspection PyAttributeOutsideInit
        self.drop_attn = layers.Dropout(self.attn_drop)

        # noinspection PyAttributeOutsideInit
        self.proj = layers.Dense(self.channels, name='proj')

        # noinspection PyAttributeOutsideInit
        self.drop_proj = layers.Dropout(self.proj_drop)

        super().build(input_shape)

    def relative_table(self, window_size):
        offset = tf.range(1 - window_size, window_size)
        offset = tf.cast(offset, self.compute_dtype)
        offset = tf.stack(tf.meshgrid(offset, offset, indexing='ij'))
        offset = tf.transpose(offset, [1, 2, 0])[None]

        window = self.window_pretrain if self.window_pretrain > 0 else window_size

        offset *= 8. / (tf.cast(window, self.compute_dtype) - 1.)
        offset = tf.sign(offset) * tf.math.log1p(tf.abs(offset)) / np.log(8)

        return offset

    def with_mask(self, attn, mask, length):
        mask_windows = tf.shape(mask)[1]
        attn = tf.reshape(attn, shape=[-1, mask_windows, self.num_heads, length, length])
        attn += mask
        attn = tf.reshape(attn, shape=[-1, self.num_heads, length, length])

        return attn

    def call(self, inputs, **kwargs):
        inputs, window_size, relative_index, attention_mask = inputs
        height, width = tf.unstack(tf.shape(inputs)[1:3])
        length = window_size ** 2

        qkv = self.qkv(inputs)
        if self.swin_v2 and self.qkv_bias:
            k_bias = tf.zeros_like(self.v_bias, self.compute_dtype)
            qkv_bias = tf.concat([self.q_bias, k_bias, self.v_bias], axis=0)
            qkv = tf.nn.bias_add(qkv, qkv_bias)

        # QKV heads partition - fused with windows partitioning
        # qkv = tf.reshape(qkv, [-1, length, 3, self.num_heads, self.channels // self.num_heads])
        # qkv = tf.transpose(qkv, [2, 0, 3, 1, 4])

        qkv = window_partition_fused(qkv, height, width, window_size, self.num_heads)

        q, k, v = tf.unstack(qkv, 3)
        if self.swin_v2:
            scale = tf.minimum(self.scale, np.log(1. / .01))
            scale = tf.exp(scale)
            q = tf.math.l2_normalize(q, axis=-1, epsilon=1.55e-5)
            k = tf.math.l2_normalize(k, axis=-1, epsilon=1.55e-5)
        else:
            scale = self.scale
        q *= scale
        attn = tf.matmul(q, k, transpose_b=True)

        if self.swin_v2:
            relative_bias = self.cpb0(self.relative_table(window_size))
            relative_bias = self.cpb1(relative_bias)
            relative_bias = tf.reshape(relative_bias, [-1, self.num_heads])
            bias = tf.gather(relative_bias, relative_index) * 16.
        else:
            bias = tf.gather(self.relative_bias, relative_index)
        bias = tf.reshape(bias, [length, length, -1])
        bias = tf.transpose(bias, perm=[2, 0, 1])
        attn = attn + bias[None]

        attn = self.with_mask(attn, attention_mask, length)

        attn = tf.nn.softmax(attn)
        attn = self.drop_attn(attn)

        outputs = tf.matmul(attn, v)

        # V heads merge - fused with windows merging
        # outputs = tf.transpose(outputs, perm=[0, 2, 1, 3])
        # outputs = tf.reshape(outputs, [-1, length, self.channels])

        outputs = window_reverse_fused(outputs, height, width, window_size, self.num_heads)

        outputs = self.proj(outputs)
        outputs = self.drop_proj(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def get_config(self):
        config = super().get_config()

        config.update({
            'num_heads': self.num_heads,
            'qkv_bias': self.qkv_bias,
            'qk_scale': self.qk_scale,
            'attn_drop': self.attn_drop,
            'proj_drop': self.proj_drop,
            'window_pretrain': self.window_pretrain,
            'swin_v2': self.swin_v2
        })

        return config
